import pathlib
from unittest.mock import patch

import botocore.exceptions
import pytest

from awspub import context, exceptions, sns

curdir = pathlib.Path(__file__).parent.resolve()


@pytest.mark.parametrize(
    "imagename,called_sns_publish, publish_call_count",
    [
        ("test-image-10", True, 1),
        ("test-image-11", True, 2),
        ("test-image-12", True, 2),
    ],
)
def test_sns_publish(imagename, called_sns_publish, publish_call_count):
    """
    Test the send_notification logic
    """
    with patch("boto3.client") as bclient_mock:
        instance = bclient_mock.return_value
        ctx = context.Context(curdir / "fixtures/config1.yaml", None)
        instance.describe_regions.return_value = {
            "Regions": [{"RegionName": "eu-central-1"}, {"RegionName": "us-east-1"}]
        }
        instance.list_buckets.return_value = {"Buckets": [{"Name": "bucket1"}]}

        sns.SNSNotification(ctx, imagename).publish()
        assert instance.publish.called == called_sns_publish
        assert instance.publish.call_count == publish_call_count


@pytest.mark.parametrize(
    "imagename",
    [
        ("test-image-10"),
        ("test-image-11"),
        ("test-image-12"),
    ],
)
def test_sns_publish_fail_with_invalid_topic(imagename):
    """
    Test the send_notification logic
    """
    with patch("boto3.client") as bclient_mock:
        instance = bclient_mock.return_value
        ctx = context.Context(curdir / "fixtures/config1.yaml", None)
        instance.describe_regions.return_value = {
            "Regions": [{"RegionName": "eu-central-1"}, {"RegionName": "us-east-1"}]
        }
        instance.list_buckets.return_value = {"Buckets": [{"Name": "bucket1"}]}

        # topic1 is invalid topic
        def side_effect(*args, **kwargs):
            topic_arn = kwargs.get("TopicArn")
            if "topic1" in topic_arn and "us-east-1" in topic_arn:
                error_reponse = {
                    "Error": {
                        "Code": "NotFoundException",
                        "Message": "An error occurred (NotFound) when calling the Publish operation: "
                        "Topic does not exist.",
                    }
                }
                raise botocore.exceptions.ClientError(error_reponse, "")

        instance.publish.side_effect = side_effect

        with pytest.raises(exceptions.AWSNotificationException):
            sns.SNSNotification(ctx, imagename).publish()


@pytest.mark.parametrize(
    "imagename",
    [
        ("test-image-10"),
        ("test-image-11"),
        ("test-image-12"),
    ],
)
def test_sns_publish_fail_with_unauthorized_user(imagename):
    """
    Test the send_notification logic
    """
    with patch("boto3.client") as bclient_mock:
        instance = bclient_mock.return_value
        ctx = context.Context(curdir / "fixtures/config1.yaml", None)
        instance.describe_regions.return_value = {
            "Regions": [{"RegionName": "eu-central-1"}, {"RegionName": "us-east-1"}]
        }
        instance.list_buckets.return_value = {"Buckets": [{"Name": "bucket1"}]}

        error_reponse = {
            "Error": {
                "Code": "AuthorizationError",
                "Message": "User are not authorized perform SNS Notification service",
            }
        }
        instance.publish.side_effect = botocore.exceptions.ClientError(error_reponse, "")

        with pytest.raises(exceptions.AWSAuthorizationException):
            sns.SNSNotification(ctx, imagename).publish()


@pytest.mark.parametrize(
    "imagename, partition, regions_in_partition, expected",
    [
        (
            "test-image-10",
            "aws-cn",
            ["cn-north1", "cn-northwest-1"],
            [],
        ),
        (
            "test-image-11",
            "aws",
            ["us-east-1", "eu-central-1"],
            [
                "arn:aws:sns:us-east-1:1234:topic1",
                "arn:aws:sns:eu-central-1:1234:topic2",
            ],
        ),
        (
            "test-image-12",
            "aws",
            ["us-east-1", "eu-central-1"],
            [
                "arn:aws:sns:us-east-1:1234:topic1",
                "arn:aws:sns:eu-central-1:1234:topic1",
            ],
        ),
    ],
)
def test_sns__get_topic_arn(imagename, partition, regions_in_partition, expected):
    """
    Test the send_notification logic
    """
    with patch("boto3.client") as bclient_mock:
        instance = bclient_mock.return_value
        ctx = context.Context(curdir / "fixtures/config1.yaml", None)
        sns_conf = ctx.conf["images"][imagename]["sns"]
        instance.describe_regions.return_value = {"Regions": [{"RegionName": r} for r in regions_in_partition]}
        instance.list_buckets.return_value = {"Buckets": [{"Name": "bucket1"}]}

        instance.get_caller_identity.return_value = {"Account": "1234", "Arn": f"arn:{partition}:iam::1234:user/test"}

        topic_arns = []
        for topic in sns_conf:
            for topic_name, topic_conf in topic.items():
                sns_regions = sns.SNSNotification(ctx, imagename)._sns_regions(topic_conf)
                for region in sns_regions:
                    res_arn = sns.SNSNotification(ctx, imagename)._get_topic_arn(topic_name, region)
                    topic_arns.append(res_arn)

        assert topic_arns == expected


@pytest.mark.parametrize(
    "imagename,regions_in_partition,regions_expected",
    [
        ("test-image-10", ["us-east-1", "eu-west-1"], {"topic1": ["us-east-1"]}),
        (
            "test-image-11",
            ["us-east-1", "eu-west-1"],
            {"topic1": ["us-east-1"], "topic2": []},
        ),
        ("test-image-12", ["eu-northwest-1", "ap-southeast-1"], {"topic1": ["eu-northwest-1", "ap-southeast-1"]}),
    ],
)
def test_sns_regions(imagename, regions_in_partition, regions_expected):
    """
    Test the regions for a given image
    """
    with patch("boto3.client") as bclient_mock:
        instance = bclient_mock.return_value
        instance.describe_regions.return_value = {"Regions": [{"RegionName": r} for r in regions_in_partition]}
        ctx = context.Context(curdir / "fixtures/config1.yaml", None)
        sns_conf = ctx.conf["images"][imagename]["sns"]
        instance.list_buckets.return_value = {"Buckets": [{"Name": "bucket1"}]}

        sns_regions = {}
        for topic in sns_conf:
            for topic_name, topic_conf in topic.items():
                sns_regions[topic_name] = sns.SNSNotification(ctx, imagename)._sns_regions(topic_conf)

        assert sns_regions == regions_expected



"""
Methods used to handle notifications for AWS using SNS
"""

import json
import logging
from typing import Any, Dict, List

import boto3
from botocore.exceptions import ClientError
from mypy_boto3_sns.client import SNSClient
from mypy_boto3_sts.client import STSClient

from awspub.common import _get_regions
from awspub.context import Context
from awspub.exceptions import AWSAuthorizationException, AWSNotificationException
from awspub.s3 import S3

logger = logging.getLogger(__name__)


class SNSNotification(object):
    """
    A data object that contains validation logic and
    structuring rules for SNS notification JSON
    """

    def __init__(self, context: Context, image_name: str):
        """
        Construct a message and verify that it is valid
        """
        self._ctx: Context = context
        self._image_name: str = image_name
        self._s3: S3 = S3(context)

    @property
    def conf(self) -> List[Dict[str, Any]]:
        """
        The sns configuration for the current image (based on "image_name") from context
        """
        return self._ctx.conf["images"][self._image_name]["sns"]

    def _sns_regions(self, topic_config: Dict[Any, Any]) -> List[str]:
        """
        Get the sns regions. Either configured in the sns configuration
        or all available regions.
        If a region is listed that is not available in the currently used partition,
        that region will be ignored (eg. having us-east-1 configured but running in the aws-cn
        partition doesn't include us-east-1 here).
        """

        regions_configured = topic_config["regions"] if "regions" in topic_config else []
        sns_regions = _get_regions(self._s3.bucket_region, regions_configured)

        return sns_regions

    def _get_topic_arn(self, topic_name: str, region_name: str) -> str:
        """
        Calculate topic ARN based on partition, region, account and topic name
        :param topic_name: Name of topic
        :type topic_name: str
        :param region_name: name of region
        :type region_name: str
        :return: return topic ARN
        :rtype: str
        """

        stsclient: STSClient = boto3.client("sts", region_name=region_name)
        resp = stsclient.get_caller_identity()

        account = resp["Account"]
        # resp["Arn"] has string format "arn:partition:iam::accountnumber:user/iam_role"
        partition = resp["Arn"].rsplit(":")[1]

        return f"arn:{partition}:sns:{region_name}:{account}:{topic_name}"

    def publish(self) -> None:
        """
        send notification to subscribers
        """

        for topic in self.conf:
            for topic_name, topic_config in topic.items():
                for region_name in self._sns_regions(topic_config):
                    snsclient: SNSClient = boto3.client("sns", region_name=region_name)
                    try:
                        snsclient.publish(
                            TopicArn=self._get_topic_arn(topic_name, region_name),
                            Subject=topic_config["subject"],
                            Message=json.dumps(topic_config["message"]),
                            MessageStructure="json",
                        )
                    except ClientError as e:
                        exception_code: str = e.response["Error"]["Code"]
                        if exception_code == "AuthorizationError":
                            raise AWSAuthorizationException(
                                "Profile does not have a permission to send the SNS notification."
                                " Please review the policy."
                            )
                        else:
                            raise AWSNotificationException(str(e))
                    logger.info(
                        f"The SNS notification {topic_config['subject']}"
                        f" for the topic {topic_name} in {region_name} has been sent."
                    )
